"""Messages for application."""

from __future__ import annotations

from dataclasses import dataclass

from par_ai_core.llm_providers import LlmProvider
from rich.console import ConsoleRenderable, RenderableType, RichCast
from textual.message import Message
from textual.message_pump import MessagePump
from textual.notifications import SeverityLevel
from textual.widgets import Input, TextArea

from parllama.messages.shared import SessionChanges
from parllama.models.ollama_data import FullModel
from parllama.utils import TabType


@dataclass
class AppRequest(Message):
    """Request to app to perform an action."""

    widget: MessagePump | None


@dataclass
class RegisterForUpdates(Message):
    """Register widget for updates."""

    widget: MessagePump
    event_names: list[str]


@dataclass
class UnRegisterForUpdates(Message):
    """Unregister widget for updates."""

    widget: MessagePump


@dataclass
class LocalModelCopied(Message):
    """Message to notify screen that local model has been copied."""

    src_model_name: str
    dst_model_name: str
    success: bool


@dataclass
class LocalModelCopyRequested(AppRequest):
    """Local model copy requested."""

    src_model_name: str
    dst_model_name: str


@dataclass
class SiteModelsLoaded(Message):
    """Message to notify screen that site models are loaded."""

    ollama_namespace: str


@dataclass
class SiteModelsRefreshRequested(AppRequest):
    """Site models refresh requested."""

    ollama_namespace: str
    force: bool


@dataclass
class StatusMessage(Message):
    """Message to update status bar."""

    msg: RenderableType
    log_it: bool = True


@dataclass
class PsMessage(Message):
    """Message to update ps status bar."""

    msg: RenderableType


@dataclass
class LocalModelListLoaded(Message):
    """Message to notify that local model list data is loaded."""


@dataclass
class LocalModelPulled(Message):
    """Message to notify screen that a model has been pulled."""

    model_name: str
    success: bool


@dataclass
class LocalModelPushed(Message):
    """Message to notify screen that a model has been pushed."""

    model_name: str
    success: bool


@dataclass
class LocalModelCreateRequested(AppRequest):
    """Message to notify that a model create has been requested."""

    model_name: str
    model_code: str
    quantization_level: str | None


@dataclass
class LocalCreateModelFromExistingRequested(AppRequest):
    """Message to open create model screen with current model file as starting point."""

    model_name: str
    model_code: str
    quantization_level: str | None


@dataclass
class LocalModelCreated(Message):
    """Message to notify that a model has been created."""

    model_name: str
    quantization_level: str | None
    model_code: str
    success: bool


@dataclass
class LocalModelPullRequested(AppRequest):
    """Message to notify that a model pull has been requested."""

    model_name: str
    notify: bool = True


@dataclass
class LocalModelPushRequested(AppRequest):
    """Message to notify that a model pull has been requested."""

    model_name: str


@dataclass
class LocalModelListRefreshRequested(AppRequest):
    """Message to notify that a local model list refresh has been requested."""


@dataclass
class LocalModelDeleted(Message):
    """Message to notify that a local model has been deleted."""

    model_name: str


@dataclass
class LocalModelDeleteRequested(AppRequest):
    """Message to notify that a local model delete has been requested."""

    model_name: str


@dataclass
class LocalModelDelete(Message):
    """Message to notify that a local model should be deleted."""

    model_name: str


@dataclass
class ShowLocalModel(Message):
    """Message to notify that a local model show has been requested."""

    model: FullModel


@dataclass
class SendToClipboard(Message):
    """Used to send a string to the clipboard."""

    message: str
    notify: bool = True


@dataclass
class SetModelNameLoading(Message):
    """Set model name loading indicator."""

    model_name: str
    loading: bool


@dataclass
class ChangeTab(Message):
    """Change to requested tab."""

    tab: TabType


@dataclass
class ModelInteractRequested(Message):
    """Message to notify that a model interact has been requested."""

    model_name: str


@dataclass
class UpdateChatControlStates(Message):
    """Notify that chat control states need to be updated."""


@dataclass
class UpdateTabLabel(Message):
    """Update tab label."""

    tab_id: str
    tab_label: str


@dataclass
class UpdateChatStatus(Message):
    """Update chat status."""


# ---------- Prompt Related Messages ---------- #


@dataclass
class PromptListChanged(Message):
    """Notify that prompt list has changed."""


@dataclass
class PromptMessage(Message):
    """Prompt base class."""

    prompt_id: str


@dataclass
class PromptDeleteRequested(AppRequest):
    """Message to notify that a prompt delete has been requested."""

    prompt_id: str


@dataclass
class DeletePrompt(PromptMessage):
    """Request prompt be deleted."""


@dataclass
class DeletePromptMessage(PromptMessage):
    """Request message be deleted from prompt."""

    message_id: str


@dataclass
class PromptSelected(PromptMessage):
    """Notify that a prompt has been selected."""

    temperature: float | None = None
    llm_provider: LlmProvider | None = None
    model_name: str | None = None


@dataclass
class PromptListLoaded(Message):
    """Prompt list loaded"""


# ---------- Session Related Messages ---------- #


@dataclass
class SessionListChanged(Message):
    """Notify that session list has changed."""


@dataclass
class SessionMessage(Message):
    """Session base class."""

    session_id: str


@dataclass
class SessionToPrompt(SessionMessage):
    """Request session be copied to prompt."""

    prompt_name: str | None = None
    submit_on_load: bool = False


@dataclass
class StopChatGeneration(SessionMessage):
    """Request chat generation to be stopped."""


@dataclass
class ChatGenerationAborted(SessionMessage):
    """Chat generation has been aborted."""


@dataclass
class ChatMessage(Message):
    """Chat message class"""

    parent_id: str
    message_id: str
    is_final: bool = False


@dataclass
class ChatMessageDeleted(Message):
    """Chat message deleted class"""

    parent_id: str
    message_id: str


@dataclass
class ChatMessageSent(SessionMessage):
    """Chat message sent class"""


@dataclass
class SessionSelected(SessionMessage):
    """Notify that session has been selected."""

    new_tab: bool = False


@dataclass
class DeleteSession(SessionMessage):
    """Request session be deleted."""


@dataclass
class NewChatSession(SessionMessage):
    """New chat session class"""


@dataclass
class SessionUpdated(SessionMessage):
    """Session Was Updated"""

    changed: SessionChanges


@dataclass
class LogIt(Message):
    """Log message."""

    msg: ConsoleRenderable | RichCast | str | object
    notify: bool = False
    severity: SeverityLevel = "information"
    timeout: int = 5


@dataclass
class ImportReady(Message):
    """Import ready message."""


@dataclass
class ToggleInputMode(Message):
    """Toggle between single and multi-line input mode."""


@dataclass
class ClearChatInputHistory(Message):
    """Clear chat history."""


# ---------- Provider Related Messages ---------- #
@dataclass
class RefreshProviderModelsRequested(AppRequest):
    """Refresh provider models."""


@dataclass
class ProviderModelsChanged(Message):
    """Provider models refreshed."""

    provider: LlmProvider | None = None


@dataclass
class ProviderModelSelected(Message):
    """Provider model selected."""

    provider: LlmProvider
    model_name: str


@dataclass
class HistoryPrev(Message):
    """Posted when the up arrow key is pressed."""

    input: Input | TextArea
    """The `Input` widget."""

    @property
    def control(self) -> Input | TextArea:
        """Alias for self.input."""
        return self.input


@dataclass
class HistoryNext(Message):
    """Posted when the down arrow key is pressed."""

    input: Input | TextArea
    """The `Input` widget."""

    @property
    def control(self) -> Input | TextArea:
        """Alias for self.input."""
        return self.input



"""Toggleable hidden text input field."""

from __future__ import annotations

from par_ai_core.utils import to_class_case
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import Button, Input

from parllama.icons import EYE_EMOJI


class HiddenInputField(Widget):
    """Toggleable hidden text input field."""

    DEFAULT_CSS = """
    HiddenInputField {
        height: 3;
        width: 1fr;
        Input {
            width: 1fr;
        }
    }
    """
    base_name: str
    input: Input

    def __init__(self, base_name: str, **kwargs) -> None:
        """Initialize the Hidden Field."""
        base_name = to_class_case(base_name.replace("/", ""))
        super().__init__(name=f"{base_name}HiddenField")

        self.base_name = base_name
        self.input = Input(id=f"{self.base_name}Input", password=True, **kwargs)

    def compose(self) -> ComposeResult:
        """Compose the list item."""
        with Horizontal(classes="width-fr-1"):
            yield self.input
            btn = Button(EYE_EMOJI, id="show", classes="field-button")
            btn.tooltip = "Show/Hide"
            yield btn

    @on(Button.Pressed, "#show")
    def show(self, event: Button.Pressed) -> None:
        """Toggle the hidden input field."""
        event.stop()
        ctrl = self.query_one(Input)
        ctrl.password = not ctrl.password

    @property
    def value(self) -> str:
        """Get the value."""
        return self.input.value

    @value.setter
    def value(self, value: str) -> None:
        """Set the value."""
        self.input.value = value



"""Session configuration widget."""

from __future__ import annotations

from par_ai_core.llm_config import LlmConfig
from par_ai_core.llm_providers import LlmProvider
from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.message import Message
from textual.widgets import Button, Input, Label, Rule, Select, Static

from parllama.chat_manager import chat_manager
from parllama.chat_session import ChatSession
from parllama.messages.messages import (
    PromptSelected,
    ProviderModelSelected,
    RegisterForUpdates,
    SessionSelected,
    SessionUpdated,
    UnRegisterForUpdates,
    UpdateChatStatus,
)
from parllama.settings_manager import settings
from parllama.widgets.input_blur_submit import InputBlurSubmit
from parllama.widgets.provider_model_select import ProviderModelSelect


class SessionConfig(VerticalScroll):
    """Session configuration widget."""

    DEFAULT_CSS = """
SessionConfig {
    width: 50;
    height: 1fr;
    dock: right;
    padding: 1;
    #session_name_input {
        width: 41;
    }
    #temperature_input {
        width: 12;
    }
    #num_ctx_input {
        width: 15;
    }
    #new_button {
        margin-left: 2;
        min-width: 9;
        background: $warning-darken-2;
        border-top: tall $warning-lighten-1;
    }
    Label {
        margin: 1;
        background: transparent;
    }
}
    """
    BINDINGS = []
    session: ChatSession
    provider_model_select: ProviderModelSelect

    def __init__(self, **kwargs) -> None:
        """Initialise the view."""
        super().__init__(**kwargs)

        session_name = chat_manager.mk_session_name("New Chat")

        self.provider_model_select = ProviderModelSelect(update_settings=True)
        self.temperature_input: InputBlurSubmit = InputBlurSubmit(
            id="temperature_input",
            value=(f"{settings.last_llm_config.temperature:.2f}"),
            max_length=4,
            restrict=r"^\d?\.?\d?\d?$",
            valid_empty=False,
        )

        self.num_ctx_input: InputBlurSubmit = InputBlurSubmit(
            id="num_ctx_input",
            value=str(settings.last_llm_config.num_ctx or 2048),
            max_length=6,
            type="integer",
            valid_empty=False,
        )

        self.session_name_input: InputBlurSubmit = InputBlurSubmit(
            id="session_name_input",
            value=session_name,
            valid_empty=False,
        )

        self.session = chat_manager.get_or_create_session(
            session_id=None,
            session_name=session_name,
            llm_config=LlmConfig(
                provider=self.provider_model_select.provider_name,
                model_name=self.provider_model_select.model_name,
                base_url=settings.provider_base_urls[self.provider_model_select.provider_name],
                temperature=self.get_temperature(),
            ),
            widget=self,
        )

    async def on_mount(self) -> None:
        """Set up the dialog once the DOM is ready."""
        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "ProviderModelsChanged",
                    "SessionSelected",
                    "SessionUpdated",
                ],
            )
        )
        self.display = settings.always_show_session_config or not self.is_valid()

    async def on_unmount(self) -> None:
        """Remove dialog from updates when unmounted."""
        self.app.post_message(UnRegisterForUpdates(widget=self))

    def compose(self) -> ComposeResult:
        """Compose the content of the view."""
        with Horizontal(classes="height-auto"):
            yield Static("Session Config", classes="width-fr-1 pt-1")
            yield Button("New", id="new_button", variant="warning")
        yield Rule()
        with Horizontal(classes="height-auto"):
            yield Label("Name")
            yield self.session_name_input
        yield self.provider_model_select
        with Horizontal(classes="height-3"):
            yield Label("Temperature")
            yield self.temperature_input
        with Horizontal(classes="height-3"):
            with Label("Max Context Size") as lbl:
                lbl.tooltip = "0 = default. Ollama default is 2048"
            yield self.num_ctx_input

    async def action_new_session(self, session_name: str = "New Chat") -> None:
        """Start new session"""
        # self.notify("New session")
        with self.prevent(Input.Changed):
            old_session = self.session
            old_session.remove_sub(self)
            self.session = chat_manager.new_session(
                session_name=session_name,
                llm_config=LlmConfig(
                    provider=LlmProvider(self.provider_model_select.provider_name),
                    model_name=self.provider_model_select.model_name,
                    temperature=self.get_temperature(),
                ),
                widget=self,
            )
            self.session_name_input.value = self.session.name
            # self.session.batching = False

    def set_model_name(self, model_name: str) -> None:
        """Set model names"""
        self.provider_model_select.set_model_name(model_name)

    @on(Button.Pressed, "#new_button")
    async def on_new_button_pressed(self, event: Button.Pressed) -> None:
        """New button pressed"""
        event.stop()
        await self.action_new_session()

    def get_temperature(self) -> float:
        """Get temperature from input field"""
        try:
            return float(self.temperature_input.value)
        except ValueError:
            return settings.last_llm_config.temperature

    @on(SessionSelected)
    def on_session_selected(self, event: SessionSelected) -> None:
        """Handle session selected event."""
        event.stop()

    @on(Input.Submitted, "#temperature_input")
    def temperature_input_changed(self, event: Message) -> None:
        """Handle temperature input change"""
        event.stop()
        if not self.temperature_input.value:
            return
        try:
            settings.last_llm_config.temperature = float(self.temperature_input.value)
        except ValueError:
            return
        self.session.temperature = settings.last_llm_config.temperature
        settings.save()

    @on(Input.Submitted, "#num_ctx_input")
    def num_ctx_input_changed(self, event: Message) -> None:
        """Handle num_ctx input change"""
        event.stop()
        if not self.num_ctx_input.value:
            return
        try:
            settings.last_llm_config.num_ctx = int(self.num_ctx_input.value)
        except ValueError:
            return
        self.session.num_ctx = settings.last_llm_config.num_ctx
        settings.save()

    @on(Input.Submitted, "#session_name_input")
    def session_name_input_changed(self, event: Input.Submitted) -> None:
        """Handle session name input change"""
        event.stop()
        event.prevent_default()
        # self.app.post_message(LogIt("CT session_name_input_changed"))
        session_name: str = self.session_name_input.value.strip()
        if not session_name:
            return
        with self.prevent(Input.Changed, Input.Submitted):
            self.session.name = chat_manager.mk_session_name(session_name)
        settings.last_chat_session_id = self.session.id
        settings.save()

    @on(ProviderModelSelected)
    def on_provider_model_selected(self, event: ProviderModelSelected) -> None:
        """Handle provider model selected event."""
        event.stop()
        self.session.llm_config.base_url = settings.provider_base_urls[event.provider]
        self.session.llm_provider_name = event.provider
        self.session.llm_model_name = event.model_name
        self.post_message(UpdateChatStatus())

    @on(SessionUpdated)
    def session_updated(self, event: SessionUpdated) -> None:
        """Handle a session updated event"""
        # self.notify(f"Session Config updated: {event.changed}")
        # Allow event to propagate to parent
        # event.stop()
        if "name" in event.changed:
            with self.prevent(Input.Changed, Input.Submitted):
                self.session_name_input.value = self.session.name

        if "temperature" in event.changed:
            with self.prevent(Input.Changed, Input.Submitted):
                self.temperature_input.value = str(self.session.temperature)

    async def load_session(self, session_id: str) -> bool:
        """Load a session"""
        # self.app.post_message(LogIt("SC load_session: " + session_id))
        session = chat_manager.get_session(session_id, self)
        if session is None:
            self.notify(f"Chat session not found: {session_id}", severity="error")
            return False
        session.load()
        old_session = self.session
        old_session.remove_sub(self)
        self.session = session

        with self.prevent(Input.Changed, Select.Changed):
            self.session_name_input.value = self.session.name
            self.provider_model_select.provider_select.value = self.session.llm_provider_name
            self.provider_model_select.provider_select_changed()
            self.provider_model_select.set_model_name(self.session.llm_model_name)
            if self.provider_model_select.model_select.value == Select.BLANK:
                self.notify("Model defined in session is not installed", severity="warning")
            self.temperature_input.value = str(self.session.temperature)

        return True

    async def load_prompt(self, event: PromptSelected) -> bool:
        """Load a session"""
        # self.app.post_message(LogIt("SC load_prompt: " + event.prompt_id))
        # self.app.post_message(
        #     LogIt(
        #         f"{event.prompt_id},{event.llm_provider_name},{event.llm_model_name},{event.temperature}"
        #     )
        # )
        prompt = chat_manager.get_prompt(event.prompt_id)
        if prompt is None:
            self.notify(f"Prompt not found: {event.prompt_id}", severity="error")
            return False
        prompt.load()
        # self.app.post_message(LogIt(f"{prompt.id},{prompt.name}"))
        old_session = self.session
        old_session.remove_sub(self)
        llm_config: LlmConfig = old_session.llm_config.clone()
        if event.temperature is not None:
            llm_config.temperature = event.temperature
        if event.llm_provider:
            llm_config.provider = event.llm_provider
        if event.model_name:
            llm_config.model_name = event.model_name
        self.session = chat_manager.new_session(
            session_name=prompt.name or old_session.name,
            llm_config=llm_config,
            widget=self,
        )
        self.provider_model_select.provider_select.value = llm_config.provider
        self.set_model_name(self.session.llm_model_name)
        # if self.provider_model_select.model_select.value == Select.BLANK:
        #     self.notify(
        #         f"Prompt model: {self.session.llm_model_name} not found",
        #         severity="warning",
        #     )
        with self.prevent(Input.Changed, Select.Changed):
            self.temperature_input.value = str(self.session.temperature)
            self.session_name_input.value = self.session.name

        return True

    def is_valid(self) -> bool:
        """Check if valid"""
        return (
            self.provider_model_select.is_valid()
            and len(str(self.temperature_input.value)) > 0
            and len(str(self.session_name_input.value)) > 0
        )



"""Session messages for par event system"""

from __future__ import annotations

from dataclasses import dataclass

from par_ai_core.llm_config import LlmConfig

from parllama.messages.shared import SessionChanges
from parllama.par_event_system import ParEventBase


@dataclass
class ParSessionMessage(ParEventBase):
    """Session message base class"""

    session_id: str


@dataclass
class ParSessionUpdated(ParSessionMessage):
    """Session was updated"""

    changed: SessionChanges


@dataclass
class ParSessionDelete(ParSessionMessage):
    """Request session be deleted."""


@dataclass
class ParSessionAutoName(ParSessionMessage):
    """Request session be auto named."""

    llm_config: LlmConfig
    context: str



"""Provides edit custom prompt dialog."""

from __future__ import annotations

from datetime import UTC, datetime

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical, VerticalScroll
from textual.events import Event
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Input, Label, Select, Static, TextArea

from parllama.chat_message import ParllamaChatMessage
from parllama.chat_prompt import ChatPrompt
from parllama.messages.messages import DeletePromptMessage
from parllama.widgets.custom_prompt_message_edit import CustomPromptMessageEdit
from parllama.widgets.field_set import FieldSet


class EditPromptDialog(ModalScreen[bool]):
    """Modal dialog that allows custom prompt editing."""

    DEFAULT_CSS = """
    EditPromptDialog {
        background: black 75%;
        align: center middle;
        &> VerticalScroll {
            background: $surface;
            width: 75%;
            height: 90%;
            min-width: 80;
            border: thick $accent;
            border-title-color: $primary;
            padding: 1;

            #last_updated {
              width: auto;
            }

            &> Input, FieldSet{
                height: 3;
                Label {
                    padding-top: 1;
                    width: 15;
                }
            }
            #add_bar {
                width: 1fr;
                height: 3;
                align: left top;
                text-align: left;
            }
            #message_container {
              width: 1fr;
              height: auto;
            }
        }
    }
    """

    BINDINGS = [
        Binding("left,up", "app.focus_previous", "", show=False),
        Binding("right,down", "app.focus_next", "", show=False),
        Binding("escape, ctrl+q", "dismiss(False)", "", show=True),
        Binding("ctrl+c", "app.copy_to_clipboard", "", show=True),
    ]
    prompt: ChatPrompt
    edit_prompt: ChatPrompt
    message_container: Vertical
    save_button: Button
    dirty: bool

    def __init__(self, prompt: ChatPrompt) -> None:
        super().__init__()
        self.prompt = prompt
        if not prompt.is_loaded:
            prompt.load()
        self.edit_prompt = self.prompt.clone()
        # we don't want the edit prompt to save or notify things
        self.edit_prompt.batching = True
        self.message_container = Vertical(id="message_container")
        self.dirty = False
        self.save_button = Button("Save", id="save")
        self.save_button.disabled = True

    def compose(self) -> ComposeResult:
        """Compose the content of the dialog."""
        with self.prevent(Input.Changed, TextArea.Changed, Select.Changed, Checkbox.Changed):
            with VerticalScroll() as vs:
                vs.border_title = "Custom Prompt Edit"
                yield self.save_button
                yield Static(f"Prompt ID: {self.edit_prompt.id}")
                yield FieldSet("Name", Input(value=self.edit_prompt.name, id="name"))
                yield FieldSet(
                    "Description",
                    Input(value=self.edit_prompt.description, id="description"),
                )
                yield FieldSet(
                    "Submit on load",
                    Checkbox(value=self.edit_prompt.submit_on_load, id="submit_on_load"),
                )
                yield FieldSet("Source", Input(value=self.edit_prompt.source, id="source"))

                yield FieldSet(
                    "Last updated",
                    Label(str(self.edit_prompt.last_updated), id="last_updated"),
                )
                with Horizontal(id="add_bar"):
                    yield Button("Add Message", id="add")

                with self.message_container:
                    for m in self.edit_prompt.messages:
                        yield CustomPromptMessageEdit(m)

    async def on_mount(self) -> None:
        """Mount the view."""
        self.query_one("#name").focus()

    @on(Button.Pressed, "#add")
    async def add_message(self, event: Button.Pressed) -> None:
        """Add a new message to the prompt."""
        event.stop()
        msg = ParllamaChatMessage(content="", role="user")
        self.edit_prompt.add_message(msg)
        me: CustomPromptMessageEdit = CustomPromptMessageEdit(msg)
        await self.message_container.mount(me)
        me.content.focus()

    @on(Button.Pressed, "#save")
    def save(self, event: Button.Pressed) -> None:
        """Copy model to create screen."""
        event.stop()
        if len(self.edit_prompt.messages) == 0:
            self.notify("Prompt must have at least one message", severity="error", timeout=5)
            return
        num_system_prompts = 0
        last_system_prompt_index = -1
        # remove empty messages and move system message to the top
        self.edit_prompt.messages = [m for m in self.edit_prompt.messages if m.content.strip()]
        for i, m in enumerate(self.edit_prompt.messages):
            if m.role == "system":
                num_system_prompts += 1
                last_system_prompt_index = i
        if num_system_prompts > 1:
            self.notify(
                "You may only have 1 system role in your prompt",
                severity="error",
                timeout=5,
            )
            return
        if num_system_prompts == 1 and last_system_prompt_index > 0:
            sp = self.edit_prompt.messages.pop(last_system_prompt_index)
            self.edit_prompt.messages.insert(0, sp)
            self.notify("System prompt moved to the top", timeout=5)

        with self.prompt.batch_changes():
            self.prompt.name = self.query_one("#name", Input).value.strip()
            self.prompt.description = self.query_one("#description", Input).value.strip()
            self.prompt.source = self.query_one("#source", Input).value.strip()
            self.prompt.submit_on_load = self.query_one("#submit_on_load", Checkbox).value
            self.prompt.replace_messages(self.edit_prompt.messages)
            self.prompt.last_updated = datetime.now(UTC)
        # self.post_message(LogIt(self.prompt))
        self.dismiss(True)

    @on(DeletePromptMessage)
    def delete_message(self, event: DeletePromptMessage) -> None:
        """Delete a message from the prompt."""
        event.stop()
        del self.edit_prompt[event.message_id]
        self.save_button.disabled = False
        self.dirty = True

    @on(Input.Changed)
    @on(Checkbox.Changed)
    @on(TextArea.Changed)
    @on(Select.Changed)
    def mark_dirty(self, event: Event) -> None:
        """Mark the prompt as dirty when a change occurs."""
        event.stop()
        self.save_button.disabled = False
        self.dirty = True



"""Streaming Chat Chunk Stats."""

from __future__ import annotations

from datetime import datetime

from pydantic import BaseModel


class TokenStats(BaseModel):
    """Streaming Chat Chunk Stats."""

    model: str
    created_at: datetime
    total_duration: int
    load_duration: int
    prompt_eval_count: int
    prompt_eval_duration: int
    eval_count: int
    eval_duration: int
    input_tokens: int
    output_tokens: int
    total_tokens: int
    time_til_first_token: int



"""A select widget for selecting a local model."""

from __future__ import annotations

from textual import on
from textual.message import Message
from textual.widgets import Select

from parllama.messages.messages import LocalModelDeleted, LocalModelListLoaded, RegisterForUpdates
from parllama.ollama_data_manager import ollama_dm
from parllama.settings_manager import settings


class LocalModelSelect(Select[str]):
    """A select widget for selecting a local model."""

    def __init__(self, **kwargs) -> None:
        """Initialise the widget."""
        self._deferred_value = None  # No deferred value.
        opts = ollama_dm.get_model_select_options()
        if "value" in kwargs:
            if len(kwargs["value"]) == 0:
                del kwargs["value"]  # Remove blank values

        if "value" in kwargs:
            if len(opts) == 0:
                if kwargs["value"]:
                    self._deferred_value = kwargs["value"]
                del kwargs[
                    "value"
                ]  # Remove the value from the kwargs to avoid conflicts with the Select widget's value attribute.
            else:
                if kwargs["value"] not in ollama_dm.get_model_names():
                    del kwargs["value"]

        super().__init__(prompt="Select Model", options=opts, **kwargs)

    async def on_mount(self) -> None:
        """Subscribe to model list changes and update the widget's options accordingly."""
        self.app.post_message(
            RegisterForUpdates(
                widget=self,
                event_names=[
                    "LocalModelDeleted",
                    "LocalModelListLoaded",
                ],
            )
        )

    @on(LocalModelListLoaded)
    @on(LocalModelDeleted)
    def on_local_model_list_loaded(self, event: Message) -> None:
        """Model list changed. Update options and preserve value if possible."""
        event.stop()
        if self._deferred_value is not None:
            old_v = self._deferred_value
            self._deferred_value = None  # Reset the deferred value.
        elif self.value != Select.BLANK:
            old_v = self.value
        elif settings.last_llm_config.model_name:
            old_v = settings.last_llm_config.model_name
        else:
            old_v = None

        opts = ollama_dm.get_model_select_options()

        if old_v is not None:
            found = False
            for _, v in opts:
                if v == old_v:
                    found = True
                    break
            if found:
                with self.prevent(Select.Changed):
                    self.set_options(opts)
                self.value = old_v
                return
        self.set_options(opts)



"""Chat messages for par event system"""

from __future__ import annotations

from dataclasses import dataclass

from parllama.par_event_system import ParEventBase


@dataclass
class ParChatMessage(ParEventBase):
    """Chat message base class"""

    parent_id: str
    message_id: str


@dataclass
class ParChatUpdated(ParChatMessage):
    """Chat message updated"""

    is_final: bool = False


@dataclass
class ParChatMessageDeleted(ParChatMessage):
    """Chat message deleted"""



"""Various icons which can be used in the application."""

from __future__ import annotations

FAVORITE_EMOJI = "🌟"
FOLDER_EMOJI = "📁"
FOLDER_OPEN_EMOJI = "📂"
GLOBE_EMOJI = "🌎"
NOTE_EMOJI = "📄"
CREDIT_CARD_EMOJI = "💳"
ID_CARD_EMOJI = "🪪"
TRASH_EMOJI = "🗑"
CLIPBOARD_EMOJI = "📋"
PENCIL_EMOJI = "✏"
MEMO_EMOJI = "📝"
PAPER_EMOJI = "📄"
BLANK_PAGE_EMOJI = "📃"
COPY_EMOJI = "❐"
GEAR_EMOJI = "⚙"
EYE_EMOJI = "👁"
LINK_EMOJI = "🔗"
HEAVY_PLUS_SIGN_EMOJI = "➕"
CROSS_MARK_EMOJI = "❌"
KEY_EMOJI = "🔑"
PERSON_EMOJI = "👤"
UNKNOWN_EMOJI = "❓"
DOWN_CHEVRON_EMOJI = "🔽"



"""Create new model view."""

from __future__ import annotations

from textual import on
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Container, Horizontal, VerticalScroll
from textual.events import Show
from textual.widgets import Button, Input, Label, TextArea

from parllama.dialogs.error_dialog import ErrorDialog
from parllama.messages.messages import ChangeTab, LocalModelCreateRequested


class ModelCreateView(Container):
    """Create new model view."""

    DEFAULT_CSS = """
    ModelCreateView {
      #name_quantize_row {
        height: 3;
        #model_name{
          width: 2fr;
          height: 3;
        }
        #ql {
          height: 3;
          width: 1fr;
          Label {
            width: 16;
            height: 1;
            margin: 1;
          }
          #quantize_level {
            width: 1fr;
            height: 3;
          }
        }
      }
      #editor {
        border: double $background;
        border-title-color: $accent;
      }
      #editor:focus {
        border: double $accent;
      }
    }
    """

    BINDINGS = [
        Binding("ctrl+c", "app.copy_to_clipboard", "", show=True),
    ]

    text_area: TextArea

    def __init__(self, **kwargs) -> None:
        """
        Initialise the screen.
        """
        super().__init__(**kwargs)
        self.name_input = Input(id="model_name", placeholder="Model Name")
        self.quantize_input = Input(
            id="quantize_level",
            placeholder="e.g. q4_0 or blank for none",
        )
        self.text_area = TextArea.code_editor("", id="editor", theme="css")
        self.text_area.indent_type = "tabs"
        self.text_area.border_title = "Model Code"
        self.create_button = Button("Create", id="create_button")

    def compose(self) -> ComposeResult:
        """Compose the content of the screen."""
        with VerticalScroll(id="main_scroll"):
            with Horizontal(id="name_quantize_row"):
                yield self.name_input
                with Horizontal(id="ql"):
                    yield Label("Quantize Level")
                    yield self.quantize_input
            yield self.text_area
            yield self.create_button

    def _on_show(self, event: Show) -> None:
        """Focus the name on show"""
        self.screen.sub_title = (  # pylint: disable=attribute-defined-outside-init
            "Create Model"
        )
        self.name_input.focus()

    @on(Button.Pressed, "#create_button")
    def action_create_model(self) -> None:
        """Create the model."""
        name = (self.name_input.value or "").strip()
        code = (self.text_area.text or "").strip()
        quantization_level = (self.quantize_input.value or "").strip()
        if not name:
            self.app.push_screen(ErrorDialog(title="Input Error", message="Please enter a model name"))
            return
        if not code:
            self.app.push_screen(ErrorDialog(title="Input Error", message="Please enter a model code"))
            return
        self.app.post_message(
            LocalModelCreateRequested(
                widget=self,
                model_name=name,
                model_code=code,
                quantization_level=quantization_level,
            )
        )
        self.app.post_message(ChangeTab(tab="Local"))



"""Various types, utility functions and decorators."""

from __future__ import annotations

from argparse import ArgumentParser, Namespace
from typing import Literal, TypeAlias

from textual.widgets import Button, Input
from textual.widgets.button import ButtonVariant

from parllama import __application_binary__, __application_title__, __version__
from parllama.icons import HEAVY_PLUS_SIGN_EMOJI, PENCIL_EMOJI, TRASH_EMOJI

DECIMAL_PRECESSION = 5

TabType: TypeAlias = Literal[
    "Local",
    "Site",
    "Chat",
    "Prompts",
    "Tools",
    "Create",
    "Options",
    #    "Secrets",
    # "Rag",
    "Logs",
]
valid_tabs: list[TabType] = [
    "Local",
    "Site",
    "Chat",
    "Prompts",
    "Tools",
    "Create",
    "Options",
    #    "Secrets",
    # "Rag",
    "Logs",
]


def mk_field_button(
    *,
    id: str,  # pylint: disable=redefined-builtin
    classes: str = "",
    emoji: str,
    tooltip: str = "",
    variant: ButtonVariant = "default",
) -> Button:
    """Make a field button."""
    btn = Button(emoji, id=id, classes=f"field-button {classes}", variant=variant)
    if tooltip:
        btn.tooltip = tooltip
    return btn


def mk_trash_button(
    *,
    id: str = "delete",  # pylint: disable=redefined-builtin
    classes: str = "",
    tooltip: str = "Delete",
) -> Button:
    """Make a trash button."""
    return mk_field_button(
        id=id,
        classes=classes,
        emoji=TRASH_EMOJI,
        tooltip=tooltip,
        variant="error",
    )


def mk_edit_button(
    *,
    id: str = "edit",  # pylint: disable=redefined-builtin
    classes: str = "",
    tooltip: str = "Edit",
) -> Button:
    """Make a edit button."""
    return mk_field_button(
        id=id,
        classes=classes,
        emoji=PENCIL_EMOJI,
        tooltip=tooltip,
    )


def mk_add_button(
    *,
    id: str = "add",  # pylint: disable=redefined-builtin
    classes: str = "",
    tooltip: str = "Add",
) -> Button:
    """Make an add button."""
    return mk_field_button(
        id=id,
        classes=classes,
        emoji=HEAVY_PLUS_SIGN_EMOJI,
        tooltip=tooltip,
    )


def clamp_input_value(input_widget: Input, min_value: int | None = None, max_value: int | None = None) -> int:
    """Clamp the value of an input widget."""
    val: int = int(input_widget.value or 0)
    if min_value is not None and val < min_value:
        input_widget.value = str(min_value)
        return min_value
    if max_value is not None and val > max_value:
        input_widget.value = str(max_value)
        return max_value
    return val


eff_word_list_cache: list[str] = []


def get_args() -> Namespace:
    """Parse and return the command line arguments.

    Returns:
            The result of parsing the arguments.
    """

    # Create the parser object.
    parser = ArgumentParser(
        prog=__application_binary__,
        description=f"{__application_title__} -- Ollama TUI.",
        epilog=f"v{__version__}",
    )

    # Add --version
    parser.add_argument(
        "-v",
        "--version",
        help="Show version information.",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    parser.add_argument(
        "-d",
        "--data-dir",
        help="Data Directory. Defaults to ~/.parllama",
    )

    parser.add_argument(
        "-u",
        "--ollama-url",
        help="URL of your Ollama instance. Defaults to http://localhost:11434",
    )

    parser.add_argument("-t", "--theme-name", help="Theme name. Defaults to par")
    parser.add_argument(
        "-m",
        "--theme-mode",
        help="Dark / Light mode. Defaults to dark",
        choices=["dark", "light"],
    )

    parser.add_argument(
        "-s",
        "--starting-tab",
        help="Starting tab. Defaults to local",
        choices=[s.lower() for s in valid_tabs],
    )
    parser.add_argument(
        "--use-last-tab-on-startup",
        help="Use last tab on startup. Defaults to 1",
        choices=["0", "1"],
    )

    parser.add_argument(
        "--load-local-models-on-startup",
        help="Load local models on startup. Defaults to 1",
        choices=["0", "1"],
    )

    parser.add_argument(
        "-p",
        "--ps-poll",
        type=int,
        help="Interval in seconds to poll ollama ps command. 0 = disable. Defaults to 3",
    )

    parser.add_argument(
        "-a",
        "--auto-name-session",
        help="Auto name session using LLM. Defaults to 0",
        choices=["0", "1"],
    )

    parser.add_argument(
        "--restore-defaults",
        help="Restore default settings and theme",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--purge-cache",
        help="Purge cached data",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--purge-chats",
        help="Purge all chat history",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--purge-prompts",
        help="Purge all custom prompts",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--no-save",
        help="Prevent saving settings for this session",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--no-chat-save",
        help="Prevent saving chats for this session",
        default=False,
        action="store_true",
    )

    # Finally, parse the command line.
    return parser.parse_args()



"""Used to edit custom prompt message"""

from __future__ import annotations

from typing import cast

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widget import Widget
from textual.widgets import Button, Select, TextArea

from parllama.chat_message import ParllamaChatMessage
from parllama.messages.messages import DeletePromptMessage
from parllama.models.ollama_data import MessageRoles, MessageRoleSelectOptions
from parllama.utils import mk_trash_button


class CustomPromptMessageEdit(Vertical):
    """Used to edit custom prompt message"""

    DEFAULT_CSS = """
    CustomPromptMessageEdit {
        width: 1fr;
        height: auto;
        border: double $accent;
        #tool_bar {
            width: 1fr;
            height: 3;
            align: right top;
            text-align: right;
        }
        Select, TextArea {
            width: 1fr;
        }
        TextArea {
            height: auto;
        }
    }
    """
    role: Select[MessageRoles]
    content: TextArea
    msg: ParllamaChatMessage

    def __init__(self, msg: ParllamaChatMessage, **kwargs) -> None:
        """Initialize the widget."""
        super().__init__(**kwargs)
        self.msg = msg
        self.role = Select(
            options=MessageRoleSelectOptions,
            value=msg.role,
        )
        self.content = TextArea(text=msg.content)

    def compose(self) -> ComposeResult:
        """Compose the child widgets."""
        with Horizontal(id="tool_bar"):
            yield mk_trash_button()
        yield self.role
        yield self.content

    @on(Select.Changed)
    def on_role_change(self) -> None:
        """Update the message role when the role select changes."""
        if not self.role or self.role.value == Select.BLANK:
            return
        self.msg.role = self.role.value  # type: ignore

    @on(TextArea.Changed)
    def on_content_change(self, event: TextArea.Changed) -> None:
        """Update the message content when the content text area changes."""
        self.msg.content = self.content.text
        if self.parent and self.parent.parent:
            cast(Widget, self.parent.parent).scroll_to_center(event.control)

    @on(Button.Pressed, "#delete")
    def on_trash_pressed(self, event: Button.Pressed) -> None:
        """Handle the trash button press."""
        event.stop()
        self.post_message(DeletePromptMessage(prompt_id="", message_id=self.msg.id))
        self.remove()
